import stactools.core
from stactools.cli.registry import Registry

stactools.core.use_fsspec()


def register_plugin(registry: Registry) -> None:
    # Register subcommands
    from stactools.sentinel1 import commands

    registry.register_subcommand(commands.create_sentinel1_command)


__version__ = "0.8.0"



import json
import logging
import os
from datetime import datetime
from typing import Any, Dict, List, Tuple

import numpy as np
import pystac
import rasterio
import rasterio.features
from pystac.extensions.raster import DataType, Sampling, Statistics
from pystac.utils import str_to_datetime
from rasterio import Affine as A
from rasterio.warp import transform_geom
from shapely.geometry import mapping, shape

logger = logging.getLogger(__name__)


class RTCMetadata:
    def __init__(self, href: str, asset: str) -> None:
        self.href = href
        self.asset = asset

        def _load_metadata_from_asset(
            scale: int = 1, precision: int = 5
        ) -> Tuple[Dict[str, Any], List[float], Dict[str, Any]]:
            """key metadata stored in Geotiff tags"""
            with rasterio.Env(
                AWS_NO_SIGN_REQUEST="YES", GDAL_DISABLE_READDIR_ON_OPEN="EMPTY_DIR"
            ):
                with rasterio.open(os.path.join(href, self.asset)) as src:
                    metadata = src.profile
                    metadata.update(src.tags())
                    metadata["SHAPE"] = src.shape

                    bbox, footprint = _get_geometries(src, scale, precision)

            return metadata, bbox, footprint

        def _get_geometries(
            src: Any, scale: int, precision: int
        ) -> Tuple[List[float], Dict[str, Any]]:
            """scale can be 1,2,4,8,16. scale=1 creates most precise footprint
            at the expense of reading all pixel values. scale=2 reads 1/4 amount
            of data be overestimates footprint by at least 1pixel (20 meters).
            """
            with rasterio.vrt.WarpedVRT(src, crs="EPSG:4326") as vrt:
                bbox = [np.round(x, decimals=precision) for x in vrt.bounds]
            arr = src.read(1, out_shape=(src.height // scale, src.width // scale))
            arr[np.where(arr != 0)] = 1
            transform = src.transform * A.scale(scale)

            # Get polygon covering entire valid data region
            rioshapes = rasterio.features.shapes(arr, transform=transform)
            max_perimeter = 0
            max_geometry = None
            for geom, val in rioshapes:
                if val == 1:
                    geometry = shape(geom)
                    if geometry.length > max_perimeter:
                        max_perimeter = geometry.length
                        max_geometry = geometry
            if max_geometry is None:
                raise ValueError("No valid footprint could be calculated")
            valid_geom = mapping(max_geometry.convex_hull)
            footprint = transform_geom(
                src.crs, "EPSG:4326", valid_geom, precision=precision
            )

            return bbox, footprint

        def _get_provenance() -> List[str]:
            """RTC products are from mosaiced GRD frames"""
            # NOTE: just GRD frame names? or additional info, like IPF from manifest.safe
            # <safe:software name="Sentinel-1 IPF" version="002.72"/>
            grd_ids = []
            for i in range(1, int(self.metadata["NUMBER_SCENES"]) + 1):
                m = json.loads(self.metadata[f"SCENE_{i}_METADATA"])
                grd_ids.append(m["title"])

            return grd_ids

        def _get_times() -> Tuple[datetime, datetime, datetime]:
            """UTC start and end times of GRDs used in RTC product"""
            times = []
            for i in range(1, int(self.metadata["NUMBER_SCENES"]) + 1):
                m = json.loads(self.metadata[f"SCENE_{i}_METADATA"])
                times += [m["start_time"], m["end_time"]]
                start = str_to_datetime(min(times))
                end = str_to_datetime(max(times))
                mid = start + (end - start) / 2
                mid = mid.replace(microsecond=0)
            return start, mid, end

        self.metadata, self.bbox, self.geometry = _load_metadata_from_asset()
        self.grd_ids = _get_provenance()
        self.start_datetime, self.datetime, self.end_datetime = _get_times()

    @property
    def product_id(self) -> str:
        date = self.metadata["DATE"].replace("-", "")
        orbNames = {"ascending": "ASC", "descending": "DSC"}
        orb = orbNames[self.metadata["ORBIT_DIRECTION"]]
        id = f"{self.metadata['MISSION_ID']}_{date}_{self.metadata['TILE_ID']}_{orb}"
        return id

    @property
    def image_media_type(self) -> str:
        return pystac.MediaType.COG

    @property
    def shape(self) -> List[int]:
        return list(self.metadata["SHAPE"])

    @property
    def image_paths(self) -> List[str]:
        return ["Gamma0_VV.tif", "Gamma0_VH.tif", "local_incident_angle.tif"]

    @property
    def absolute_orbit(self) -> int:
        return int(self.metadata["ABSOLUTE_ORBIT_NUMBER"])

    @property
    def relative_orbit(self) -> int:
        """https://forum.step.esa.int/t/sentinel-1-relative-orbit-from-filename/7042"""
        satellite_lookup = {"S1B": 27, "S1A": 73}
        modifier = satellite_lookup[self.metadata["MISSION_ID"]]
        rel_orbit = ((self.absolute_orbit - modifier) % 175) + 1
        return rel_orbit

    @property
    def orbit_state(self) -> str:
        return str(self.metadata["ORBIT_DIRECTION"])

    @property
    def platform(self) -> str:
        platformMap = dict(S1A="sentinel-1a", S1B="sentinel-1b")
        return platformMap[self.metadata["MISSION_ID"]]

    @property
    def epsg(self) -> int:
        return int(self.metadata["crs"].to_epsg())

    @property
    def valid_percent(self) -> float:
        # Common to 3 assets of RTC item
        return float(self.metadata["VALID_PIXEL_PERCENT"])

    @property
    def metadata_dict(self) -> Dict[str, Any]:
        """match s2 l2a cogs from https://earth-search.aws.element84.com/v0"""
        sentinel_metadata = {
            # NOTE: once PySTAC adds MGRS extension can move this to stac.py
            "mgrs:utm_zone": self.metadata["TILE_ID"][:2],
            "mgrs:latitude_band": self.metadata["TILE_ID"][2],
            "mgrs:grid_square": self.metadata["TILE_ID"][3:],
            "sentinel:mgrs": self.metadata["TILE_ID"],
            "sentinel:product_ids": self.grd_ids,
        }
        return sentinel_metadata

    @property
    def asset_dict(self) -> Dict[str, Any]:
        """map image_path (geotif) to pystac.Asset fields"""
        asset_dict = {
            "Gamma0_VV.tif": dict(
                key="gamma0_vv",
                title="Gamma0 VV backscatter",
                roles=["data", "gamma0"],
                raster=dict(
                    nodata=0,
                    sampling=Sampling.AREA,
                    data_type=DataType.FLOAT32,
                    statistics=Statistics({"valid_percent": self.valid_percent}),
                ),
            ),
            "Gamma0_VH.tif": dict(
                key="gamma0_vh",
                title="Gamma0 VH backscatter",
                roles=["data", "gamma0"],
                raster=dict(
                    nodata=0,
                    sampling=Sampling.AREA,
                    data_type=DataType.FLOAT32,
                    statistics=Statistics({"valid_percent": self.valid_percent}),
                ),
            ),
            "local_incident_angle.tif": dict(
                key="incidence",
                title="Local incidence angle",
                roles=["data", "local-incidence-angle"],
                raster=dict(
                    nodata=0,
                    sampling=Sampling.AREA,
                    data_type=DataType.UINT16,
                    unit="degrees",
                    scale=0.01,
                    statistics=Statistics({"valid_percent": self.valid_percent}),
                ),
            ),
        }

        return asset_dict



import json
import logging
import os

import click

from stactools.sentinel1.rtc.stac import create_collection, create_item

logger = logging.getLogger(__name__)


@click.group("rtc")
def rtc_cmd() -> None:
    """Commands for working with sentinel1 RTC data"""
    pass


@rtc_cmd.command(
    "create-collection", short_help="Create a Sentinel1 RTC STAC Collection"
)
@click.argument("destination")
def create_collection_command(destination: str) -> None:
    """Creates a STAC Collection for Sentinel1 RTC products"""

    collection = create_collection()
    json_path = os.path.join(destination, "sentinel1-rtc.json")
    collection.set_self_href(os.path.join(os.path.basename(json_path)))
    collection.validate()
    collection.save_object(dest_href=json_path)


@rtc_cmd.command(
    "create-item", short_help="Convert a Sentinel1 RTC product into a STAC item"
)
@click.argument("src")
@click.argument("dst")
@click.option(
    "-a",
    "--asset",
    default="local_incident_angle.tif",
    help="Asset geotiff metadata to read",
)
@click.option(
    "-p",
    "--providers",
    help="Path to JSON file containing array of additional providers",
)
@click.option(
    "-m",
    "--metadata",
    is_flag=True,
    default=False,
    help="Include links to original GRD metadata as STAC Assets",
)
def create_item_command(
    src: str, dst: str, asset: str, providers: str, metadata: bool
) -> None:
    """Creates a STAC Item for a given Sentinel1 RTC product

    SRC is the path to the granule
    DST is directory that a STAC Item JSON file will be created
    in. for example, ./S1B_20161121_12SYJ_ASC/S1B_20161121_12SYJ_ASC.json
    """
    additional_providers = None
    if providers is not None:
        with open(providers) as f:
            additional_providers = json.load(f)

    item = create_item(
        src,
        asset,
        additional_providers=additional_providers,
        include_grd_metadata=metadata,
    )
    item_path = os.path.join(dst, f"{item.id}.json")
    item.set_self_href(item_path)
    item.save_object()



import logging
import os

import click

from ..formats import Format
from .stac import create_collection, create_item

logger = logging.getLogger(__name__)


@click.group("slc")
def slc_cmd() -> None:
    """Commands for working with sentinel1 SLC data"""
    pass


@slc_cmd.command(
    "create-collection", short_help="Create a Sentinel1 SLC STAC Collection"
)
@click.argument("destination")
def create_collection_command(destination: str) -> None:
    """Creates a STAC Collection for Sentinel1 SLC products"""

    json_path = os.path.join(destination, "sentinel1-slc.json")
    collection = create_collection(json_path)
    collection.validate()
    collection.save_object(dest_href=json_path)


@slc_cmd.command(
    "create-item",
    short_help="Convert a Sentinel1 SLC scene into a STAC item",
)
@click.argument("src")
@click.argument("dst")
@click.option("--format", default="SAFE", type=str, help="SAFE or COG format")
def create_item_command(src: str, dst: str, format: str = "SAFE") -> None:
    """Creates a STAC Collection

    Args:
        src (str): path to the scene
        dst (str): path to the STAC Item JSON file that will be created
        format (str): Specifying the format of the granule. Currently supported formats
            are SAFE (default) and COG.
    """
    if format == "COG":
        archive_format = Format.COG
    else:
        archive_format = Format.SAFE

    item = create_item(src, archive_format=archive_format)

    item_path = os.path.join(dst, "{}.json".format(item.id))
    item.set_self_href(item_path)

    item.save_object()



from datetime import datetime
from itertools import product
from typing import Any, Dict

import pystac
from pystac import Extent, SpatialExtent, TemporalExtent
from pystac.extensions import sar, sat
from pystac.extensions.item_assets import AssetDefinition
from pystac.link import Link
from pystac.utils import str_to_datetime

from ..constants import (
    ACQUISITION_MODES,
    INSPIRE_METADATA_ASSET_KEY,
    PRODUCT_METADATA_ASSET_KEY,
    SAFE_MANIFEST_ASSET_KEY,
    SENTINEL_CONSTELLATION,
    SENTINEL_LICENSE,
    SENTINEL_PLATFORMS,
    SENTINEL_POLARIZATIONS,
    SENTINEL_PROVIDER,
)

__all__ = [
    "INSPIRE_METADATA_ASSET_KEY",
    "SAFE_MANIFEST_ASSET_KEY",
    "PRODUCT_METADATA_ASSET_KEY",
    "SENTINEL_LICENSE",
    "SENTINEL_PLATFORMS",
    "ACQUISITION_MODES",
    "SENTINEL_CONSTELLATION",
    "SENTINEL_PROVIDER",
    "SENTINEL_POLARIZATIONS",
    "SENTINEL_SLC_DESCRIPTION",
    "SENTINEL_SLC_START",
    "SENTINEL_SLC_EXTENT",
    "SENTINEL_SLC_TECHNICAL_GUIDE",
    "SENTINEL_SLC_LICENSE",
    "SENTINEL_SLC_KEYWORDS",
    "SENTINEL_SLC_SAT",
    "SENTINEL_SLC_SAR",
    "SENTINEL_SLC_SWATHS",
    "SENTINEL_SLC_ASSETS",
    "SENTINEL_SLC_IW_TPRE",
    "SENTINEL_SLC_IW_TBEAM",
    "SENTINEL_SLC_IW_TORB",
    "SENTINEL_SLC_EW_TPRE",
    "SENTINEL_SLC_EW_TBEAM",
    "SENTINEL_SLC_EW_TORB",
]

SENTINEL_SLC_DESCRIPTION = (
    "Level-1 Single Look Complex (SLC) products are images in the slant range by azimuth imaging plane, in the image plane of satellite data acquisition. Each image pixel is represented by a complex (I and Q) magnitude value and therefore contains both amplitude and phase information. Each I and Q value "  # noqa: E501
    "is 16 bits per pixel. The processing for all SLC products results in a single look in each dimension using the full available signal bandwidth. The imagery is geo-referenced using orbit and attitude data from the satellite. SLC images are produced in a zero Doppler geometry. This convention is common "  # noqa: E501
    "with the standard slant range products available from other SAR sensors."
)

SENTINEL_SLC_START: datetime = str_to_datetime("2014-10-10T00:00:00Z")
SENTINEL_SLC_EXTENT = Extent(
    SpatialExtent([-180.0, -90.0, 180.0, 90.0]),
    TemporalExtent([[SENTINEL_SLC_START, None]]),
)

SENTINEL_SLC_TECHNICAL_GUIDE = Link(
    title="Sentinel-1 Single Look Complex (SLC) Technical Guide",
    rel="about",
    target="https://sentinels.copernicus.eu/web/sentinel/technical-guides/sentinel-1-sar/products-algorithms/level-1-algorithms/single-look-complex",  # noqa: E501
)

SENTINEL_SLC_LICENSE = Link(
    title="Sentinel License",
    rel="license",
    target="https://scihub.copernicus.eu/twiki/do/view/SciHubWebPortal/TermsConditions",
)

SENTINEL_SLC_KEYWORDS = ["ground", "sentinel", "copernicus", "esa", "sar"]

SENTINEL_SLC_SAT = {
    "orbit_state": [sat.OrbitState.ASCENDING, sat.OrbitState.DESCENDING]
}

SENTINEL_SLC_SAR: Dict[str, Any] = {
    "looks_range": [1],
    "product_type": ["SLC"],
    "looks_azimuth": [1],
    "polarizations": [
        sar.Polarization.HH,
        sar.Polarization.VV,
        sar.Polarization.HV,
        sar.Polarization.VH,
        [
            sar.Polarization.HH,
            sar.Polarization.HV,
        ],
        [
            sar.Polarization.VV,
            sar.Polarization.VH,
        ],
    ],
    "frequency_band": [sar.FrequencyBand.C],
    "instrument_mode": ["IW", "EW", "SM", "WV"],
    "center_frequency": [5.405],
    "resolution_range": [
        1.7,
        2.0,
        2.5,
        2.7,
        3.1,
        3.3,
        3.6,
        3.5,
        7.9,
        9.9,
        11.6,
        13.3,
        14.4,
    ],
    "resolution_azimuth": [
        3.9,
        4.9,
        22.5,
        22.6,
        22.7,
        43.7,
        44.3,
        45.2,
        45.6,
        44.0,
    ],
    "pixel_spacing_range": [
        1.5,
        1.8,
        2.2,
        2.3,
        2.6,
        2.9,
        3.1,
        5.9,
    ],
    "pixel_spacing_azimuth": [
        3.5,
        3.6,
        4.1,
        4.2,
        14.1,
        19.9,
    ],
    "observation_direction": [sar.ObservationDirection.RIGHT],
    "looks_equivalent_number": [1],
}


SENTINEL_SLC_SWATHS = [
    "IW1",
    "IW2",
    "IW3",
    "EW1",
    "EW2",
    "EW3",
    "EW4",
    "EW5",
    "S1",
    "S2",
    "S3",
    "S4",
    "S5",
    "S6",
    "WV1",
    "WV2",
]


SENTINEL_SLC_IMAGE_ASSET_DEFINITIONS = {
    f"{swath.lower()}-{pol.lower()}": AssetDefinition(
        {
            "title": f"{swath.upper()} {pol.upper()} Data",
            "type": pystac.MediaType.COG,
            "description": (
                f"{swath.upper()} {pol.upper()} polarization backscattering"
                "coefficient, 16-bit DN."
            ),
            "roles": ["data"],
        }
    )
    for swath, pol in product(SENTINEL_SLC_SWATHS, SENTINEL_POLARIZATIONS.keys())
}

SENTINEL_SLC_SCHEMA_CALIBRATION_ASSET_DEFINITIONS = {
    f"schema-calibration-{swath.lower()}-{pol.lower()}": AssetDefinition(
        {
            "title": f"{pol.upper()} Calibration Schema",
            "type": pystac.MediaType.XML,
            "description": (
                "Calibration metadata including calibration information and the beta nought, "
                "sigma nought, gamma and digital number look-up tables that can be used for "
                "absolute product calibration."
            ),
            "roles": ["metadata"],
        }
    )
    for swath, pol in product(SENTINEL_SLC_SWATHS, SENTINEL_POLARIZATIONS.keys())
}

SENTINEL_SLC_SCHEMA_NOISE_ASSET_DEFINITIONS = {
    f"schema-noise-{swath.lower()}-{pol.lower()}": AssetDefinition(
        {
            "title": f"{pol.upper()} Noise Schema",
            "type": pystac.MediaType.XML,
            "description": "Estimated thermal noise look-up tables",
            "roles": ["metadata"],
        }
    )
    for swath, pol in product(SENTINEL_SLC_SWATHS, SENTINEL_POLARIZATIONS.keys())
}

SENTINEL_SLC_SCHEMA_PRODUCT_ASSET_DEFINITIONS = {
    f"schema-product-{swath.lower()}-{pol.lower()}": AssetDefinition(
        {
            "title": f"{pol.upper()} Product Schema",
            "type": pystac.MediaType.XML,
            "description": (
                "Describes the main characteristics corresponding to the band: state of the "
                "platform during acquisition, image properties, Doppler information, geographic "
                "location, etc."
            ),
            "roles": ["metadata"],
        }
    )
    for swath, pol in product(SENTINEL_SLC_SWATHS, SENTINEL_POLARIZATIONS.keys())
}

SENTINEL_SLC_ASSETS = {
    **SENTINEL_SLC_IMAGE_ASSET_DEFINITIONS,
    **SENTINEL_SLC_SCHEMA_CALIBRATION_ASSET_DEFINITIONS,
    **SENTINEL_SLC_SCHEMA_NOISE_ASSET_DEFINITIONS,
    **SENTINEL_SLC_SCHEMA_PRODUCT_ASSET_DEFINITIONS,
    "safe-manifest": AssetDefinition(
        {
            "title": "Manifest File",
            "type": pystac.MediaType.XML,
            "description": (
                "General product metadata in XML format. Contains a high-level textual "
                "description of the product and references to all of product's components, "
                "the product metadata, including the product identification and the resource "
                "references, and references to the physical location of each component file "
                "contained in the product."
            ),
            "roles": ["metadata"],
        }
    ),
    "thumbnail": AssetDefinition(
        {
            "title": "Preview Image",
            "type": pystac.MediaType.PNG,
            "description": (
                "An averaged, decimated preview image in PNG format. Single polarization "
                "products are represented with a grey scale image. Dual polarization products "
                "are represented by a single composite colour image in RGB with the red channel "
                "(R) representing the  co-polarization VV or HH), the green channel (G) "
                "represents the cross-polarization (VH or HV) and the blue channel (B) "
                "represents the ratio of the cross an co-polarizations."
            ),
            "roles": ["thumbnail"],
        }
    ),
}


SENTINEL_SLC_IW_TPRE = 2.299849  # Preamble length
SENTINEL_SLC_IW_TBEAM = 2.758273  # Beam cycle time
SENTINEL_SLC_IW_TORB = 12 * 86400 / 175  # Nominal orbit duration

SENTINEL_SLC_EW_TPRE = 2.299970  # Preamble length
SENTINEL_SLC_EW_TBEAM = 3.038376  # Beam cycle time
SENTINEL_SLC_EW_TORB = 12 * 86400 / 175  # Nominal orbit duration



import os
from typing import List, Tuple

import pystac

from ..metadata_links import MetadataLinks, extract_properties


def get_swath_and_polarisation(href: str, upper: bool = True) -> Tuple[str, str]:
    swath, polarisation = extract_properties(href, ["swath", "polarisation"])
    if upper:
        return (swath.upper(), polarisation.upper())
    else:
        return (swath.lower(), polarisation.lower())


class SLCMetadataLinks(MetadataLinks):
    @property
    def annotation_hrefs(self) -> List[Tuple[str, str]]:
        return [
            (
                "schema-product-{}-{}".format(*get_swath_and_polarisation(x, False)),
                os.path.join(self.granule_href, self.map_filename(x)),
            )
            for x in self.grouped_hrefs["annotation"]
            if x.endswith("xml")
        ]

    @property
    def calibration_hrefs(self) -> List[Tuple[str, str]]:
        return [
            (
                "schema-calibration-{}-{}".format(
                    *get_swath_and_polarisation(x, False)
                ),
                os.path.join(self.granule_href, self.map_filename(x)),
            )
            for x in self.grouped_hrefs["calibration_calibration"]
        ]

    @property
    def noise_hrefs(self) -> List[Tuple[str, str]]:
        return [
            (
                "schema-noise-{}-{}".format(*get_swath_and_polarisation(x, False)),
                os.path.join(self.granule_href, self.map_filename(x)),
            )
            for x in self.grouped_hrefs["calibration_noise"]
        ]

    def create_product_asset(self) -> List[Tuple[str, pystac.asset.Asset]]:
        assets = []
        desc = (
            "Describes the main characteristics corresponding to the band: state of the "
            "platform during acquisition, image properties, Doppler information, geographic "
            "location, etc."
        )
        for key, href in self.annotation_hrefs:
            # Extract polarisation from href
            swath, polarisation = get_swath_and_polarisation(href)
            if polarisation:
                # Add polarisation to title
                title = f"{swath} {polarisation} Product Schema"
                asset = pystac.Asset(
                    href=href,
                    media_type=pystac.MediaType.XML,
                    title=title,
                    roles=["metadata"],
                    description=desc,
                )
                assets.append((key, asset))
        return assets

    def create_calibration_asset(self) -> List[Tuple[str, pystac.asset.Asset]]:
        assets = []
        desc = (
            "Calibration metadata including calibration information and the beta nought, "
            "sigma nought, gamma and digital number look-up tables that can be used for "
            "absolute product calibration."
        )
        for key, href in self.calibration_hrefs:
            # Extract polarisation from href
            swath, polarisation = get_swath_and_polarisation(href)
            if polarisation:
                # Add polarisation to title
                title = f"{swath} {polarisation} Calibration Schema"
                asset = pystac.Asset(
                    href=href,
                    media_type=pystac.MediaType.XML,
                    title=title,
                    roles=["metadata"],
                    description=desc,
                )
                assets.append((key, asset))
        return assets

    def create_noise_asset(self) -> List[Tuple[str, pystac.asset.Asset]]:
        assets = []
        for key, href in self.noise_hrefs:
            # Extract polarisation from href
            swath, polarisation = get_swath_and_polarisation(href)
            if polarisation:
                # Add polarisation to title
                title = f"{swath} {polarisation} Noise Schema"
                asset = pystac.Asset(
                    href=href,
                    media_type=pystac.MediaType.XML,
                    title=title,
                    roles=["metadata"],
                    description="Estimated thermal noise look-up tables",
                )
                assets.append((key, asset))
        return assets



import logging
import os
from typing import Any, Optional

import pystac
import shapely
from pystac import Summaries
from pystac.extensions.eo import EOExtension
from pystac.extensions.item_assets import ItemAssetsExtension
from pystac.extensions.projection import ProjectionExtension
from pystac.extensions.sar import SarExtension
from pystac.extensions.sat import SatExtension
from stactools.core.io import ReadHrefModifier
from stactools.core.projection import transform_from_bbox

from ..bands import image_asset_from_href
from ..formats import Format
from ..metadata_links import MetadataLinks
from ..product_metadata import get_shape
from . import constants as c
from .product_metadata import GRDProductMetadata
from .properties import fill_sar_properties, fill_sat_properties

logger = logging.getLogger(__name__)


def create_collection(json_path: str) -> pystac.Collection:
    """Creates a STAC Collection for Sentinel-1 RTC"""
    # Lists of all possible values for items
    summary_dict = {
        "constellation": [c.SENTINEL_CONSTELLATION],
        "platform": c.SENTINEL_PLATFORMS,
    }

    collection = pystac.Collection(
        id="sentinel1-grd",
        description=c.SENTINEL_GRD_DESCRIPTION,
        extent=c.SENTINEL_GRD_EXTENT,
        title="Sentinel-1 GRD",
        href=json_path,
        stac_extensions=[
            SarExtension.get_schema_uri(),
            SatExtension.get_schema_uri(),
            EOExtension.get_schema_uri(),
        ],
        keywords=c.SENTINEL_GRD_KEYWORDS,
        providers=[c.SENTINEL_PROVIDER, c.SENTINEL_GRD_PROVIDER],
        summaries=Summaries(summary_dict),
    )

    # Links
    collection.links.append(c.SENTINEL_GRD_LICENSE)
    collection.links.append(c.SENTINEL_GRD_TECHNICAL_GUIDE)

    # SAR Extension
    sar = SarExtension.summaries(collection, add_if_missing=True)
    sar.looks_range = c.SENTINEL_GRD_SAR["looks_range"]
    sar.product_type = c.SENTINEL_GRD_SAR["product_type"]
    sar.looks_azimuth = c.SENTINEL_GRD_SAR["looks_azimuth"]
    sar.polarizations = c.SENTINEL_GRD_SAR["polarizations"]
    sar.frequency_band = c.SENTINEL_GRD_SAR["frequency_band"]
    sar.instrument_mode = c.SENTINEL_GRD_SAR["instrument_mode"]
    sar.center_frequency = c.SENTINEL_GRD_SAR["center_frequency"]
    sar.resolution_range = c.SENTINEL_GRD_SAR["resolution_range"]
    sar.resolution_azimuth = c.SENTINEL_GRD_SAR["resolution_azimuth"]
    sar.pixel_spacing_range = c.SENTINEL_GRD_SAR["pixel_spacing_range"]
    sar.observation_direction = c.SENTINEL_GRD_SAR["observation_direction"]
    sar.pixel_spacing_azimuth = c.SENTINEL_GRD_SAR["pixel_spacing_azimuth"]
    sar.looks_equivalent_number = c.SENTINEL_GRD_SAR["looks_equivalent_number"]

    # Satellite Extension
    sat = SatExtension.summaries(collection, add_if_missing=True)
    sat.orbit_state = c.SENTINEL_GRD_SAT["orbit_state"]

    # Item Asset Extension
    assets = ItemAssetsExtension.ext(collection, add_if_missing=True)
    assets.item_assets = c.SENTINEL_GRD_ASSETS

    return collection


def create_item(
    granule_href: str,
    read_href_modifier: Optional[ReadHrefModifier] = None,
    archive_format: Format = Format.SAFE,
    **kwargs: Any,
) -> pystac.Item:
    """Create a STC Item from a Sentinel-1 GRD scene.

    Args:
        granule_href (str): The HREF to the granule.
            This is expected to be a path to a SAFE archive (see format for other options).
        read_href_modifier: A function that takes an HREF and returns a modified HREF.
            This can be used to modify a HREF to make it readable, e.g. appending
            an Azure SAS token or creating a signed URL.
        archive_format: An enum specifying the format of the granule. Currently supported formats
            are SAFE (default) and COG.


    Returns:
        pystac.Item: An item representing the Sentinel-1 GRD scene.
    """

    metalinks = MetadataLinks(
        granule_href,
        read_href_modifier,
        archive_format,
        **kwargs,
    )

    product_metadata = GRDProductMetadata(
        metalinks.product_metadata_href,
        metalinks.grouped_hrefs,
        metalinks.map_filename,
        metalinks.manifest,
    )

    scene_id = product_metadata.scene_id

    # Remove the last segment of the scene id so the same scene reprocessed
    # at different times will have the same Item ID
    item_id = scene_id[:-5]

    item = pystac.Item(
        id=item_id,
        geometry=product_metadata.geometry,
        bbox=product_metadata.bbox,
        datetime=product_metadata.get_datetime,
        properties={},
        stac_extensions=[],
    )

    # ---- Add Extensions ----
    # SAR Extension
    sar = SarExtension.ext(item, add_if_missing=True)
    fill_sar_properties(sar, metalinks.manifest, product_metadata.resolution)

    # Satellite Extension
    sat = SatExtension.ext(item, add_if_missing=True)
    fill_sat_properties(sat, metalinks.manifest)

    # eo
    EOExtension.ext(item, add_if_missing=True)

    # Projection Extension
    projection = ProjectionExtension.ext(item, add_if_missing=True)
    projection.epsg = 4326
    projection.bbox = product_metadata.bbox
    shape = get_shape(metalinks, read_href_modifier, **kwargs)
    projection.shape = shape
    projection.transform = transform_from_bbox(projection.bbox, shape)
    centroid = shapely.geometry.shape(item.geometry).centroid
    projection.centroid = {"lat": round(centroid.y, 5), "lon": round(centroid.x, 5)}

    # --Common metadata--
    item.common_metadata.providers = [c.SENTINEL_PROVIDER]
    item.common_metadata.platform = (
        product_metadata.platform.lower()
        if product_metadata.platform
        else product_metadata.platform
    )
    item.common_metadata.constellation = c.SENTINEL_CONSTELLATION

    # Add s1 properties
    item.properties.update(product_metadata.metadata_dict)

    item.properties["s1:shape"] = shape
    item.properties["s1:product_identifier"] = scene_id

    if pdt := metalinks.manifest.find_attr(
        "stop", ".//safe:processing[@name='GRD Post Processing']"
    ):
        item.properties["s1:processing_datetime"] = f"{pdt}Z"

    # Add assets to item
    item.add_asset(*metalinks.create_manifest_asset())

    # Annotations for bands
    for asset_obj in metalinks.create_product_asset():
        item.add_asset(asset_obj[0], asset_obj[1])

    # Calibrations for bands
    for asset_obj in metalinks.create_calibration_asset():
        item.add_asset(asset_obj[0], asset_obj[1])

    # Noise for bands
    for asset_obj in metalinks.create_noise_asset():
        item.add_asset(asset_obj[0], asset_obj[1])

    # Thumbnail
    if metalinks.thumbnail_href is not None:
        desc = (
            "An averaged, decimated preview image in PNG format. Single polarization "
            "products are represented with a grey scale image. Dual polarization products "
            "are represented by a single composite colour image in RGB with the red channel "
            "(R) representing the  co-polarization VV or HH), the green channel (G) "
            "represents the cross-polarization (VH or HV) and the blue channel (B) "
            "represents the ratio of the cross an co-polarizations."
        )
        item.add_asset(
            "thumbnail",
            pystac.Asset(
                href=metalinks.thumbnail_href,
                media_type=pystac.MediaType.PNG,
                roles=["thumbnail"],
                title="Preview Image",
                description=desc,
            ),
        )

    images_media_type = None
    if archive_format == Format.SAFE:
        images_media_type = pystac.MediaType.GEOTIFF
    elif archive_format == Format.COG:
        images_media_type = pystac.MediaType.COG

    image_assets = dict(
        [
            image_asset_from_href(
                os.path.join(granule_href, image_path),
                item,
                media_type=images_media_type,
            )
            for image_path in product_metadata.image_paths
        ]
    )

    for key, asset in image_assets.items():
        assert key not in item.assets
        item.add_asset(key, asset)

    # --Links--
    item.links.append(c.SENTINEL_GRD_LICENSE)
    item.links.append(c.SENTINEL_GRD_TECHNICAL_GUIDE)

    return item



import itertools
import json
import os
import re
from typing import Any, Dict, List, Optional, Tuple

import pystac
import stactools.core.io
from lxml import etree
from stactools.core.io import ReadHrefModifier
from stactools.core.io.xml import XmlElement

from .formats import Format
from .grd.constants import SAFE_MANIFEST_ASSET_KEY


class ManifestError(Exception):
    pass


dataset_naming_pattern = re.compile(
    "^.*"
    + "(?P<mission>s1a|s1b)"
    + "-(?P<swath>s[1-6]|iw[1-3]?|ew[1-5]?|wv[1-2]|en|n[1-6]|is[1-7])"
    + "-(?P<type>slc|grd)"
    + "-(?P<polarisation>hh|hv|vv|vh)"
    + "-([0-9]{8}t[0-9]{6})"
    + "-([0-9]{8}t[0-9]{6})"
    + "-([0-9]{6})"
    + "-([0-9a-f]{6})"
    + ".*$"
)


def extract_polarisation(href: str) -> str:
    if match := dataset_naming_pattern.match(href):
        return match.group("polarisation").upper()
    elif match := re.search(r"-(hh|hv|vv|vh)\.(?:tiff|xml)$", href):
        return match.group(1).upper()
    raise RuntimeError(f"Failed to match polarisation: {href}")


def extract_properties(href: str, properties: List[str]) -> List[str]:
    matches = dataset_naming_pattern.match(href)
    if matches is not None:
        return [matches.group(name) for name in properties]
    else:
        raise RuntimeError(f"href doesn't match dataset naming pattern: {href}")


def group_files(hrefs: List[str]) -> Dict[str, List[str]]:
    def determine_group(href: str) -> str:
        if href.startswith("annotation/calibration/calibration"):
            return "calibration_calibration"
        elif href.startswith("annotation/calibration/noise"):
            return "calibration_noise"
        elif href.startswith("annotation"):
            return "annotation"
        elif href.startswith("measurement"):
            return "measurement"
        elif href.startswith("S1"):
            return "other"
        else:
            return "other_short"

    list.sort(hrefs, key=determine_group)

    grouped_hrefs = {}
    for k, v in itertools.groupby(hrefs, determine_group):
        grouped_hrefs[k] = list(v)

    return grouped_hrefs


class MetadataLinks:
    def __init__(
        self,
        granule_href: str,
        read_href_modifier: Optional[ReadHrefModifier] = None,
        archive_format: Format = Format.SAFE,
        **kwargs: Any,
    ) -> None:
        self.granule_href = granule_href
        self.href = os.path.join(granule_href, "manifest.safe")
        self.archive_format = archive_format

        self.manifest = XmlElement.from_file(self.href, read_href_modifier, **kwargs)
        data_object_section = self.manifest.find("dataObjectSection")
        if data_object_section is None:
            raise ManifestError(
                f"Manifest at {self.href} does not have a dataObjectSection"
            )

        self._data_object_section = data_object_section
        self.product_metadata_href = os.path.join(granule_href, "manifest.safe")

        if archive_format == Format.COG:
            self.product_info_href = os.path.join(granule_href, "productInfo.json")
            self.product_info = json.loads(
                stactools.core.io.read_text(
                    self.product_info_href, read_href_modifier, **kwargs
                )
            )
            self.filename_map = self.product_info["filenameMap"]

        file_location_list = self._data_object_section.findall(
            "dataObject/byteStream/fileLocation[@href]"
        )

        def href_finder(el: XmlElement) -> Optional[str]:
            href = el.find_attr("href", ".")
            if href is not None:
                return href.strip("./")
            else:
                raise RuntimeError(
                    f"No href found in XML element {etree.tostring(el.element)}"
                )

        optional_href_list = [href_finder(href) for href in file_location_list]
        href_list = [x for x in optional_href_list if x is not None]

        self.grouped_hrefs = group_files(href_list)

    def map_filename(self, filename: str) -> str:
        if self.archive_format == Format.SAFE:
            return filename
        elif self.archive_format == Format.COG:
            return str(self.filename_map[filename])
        else:
            raise RuntimeError(f"Unknown format encountered: {self.archive_format}")

    def _find_href(self, xpaths: List[str]) -> Optional[str]:
        file_path = None
        for xpath in xpaths:
            file_path = self._data_object_section.find_attr("href", xpath)
            if file_path is not None:
                break

        if file_path is None:
            return None
        else:
            # Remove relative prefix that some paths have
            file_path = file_path.strip("./")
            return os.path.join(self.granule_href, file_path)

    @property
    def thumbnail_href(self) -> Optional[str]:
        preview = os.path.join(self.granule_href, "preview")
        return os.path.join(preview, "quick-look.png")

    @property
    def annotation_hrefs(self) -> List[Tuple[str, str]]:
        return [
            (
                "schema-product-{}".format(*extract_properties(x, ["polarisation"])),
                os.path.join(self.granule_href, self.map_filename(x)),
            )
            for x in self.grouped_hrefs["annotation"]
            if x.endswith("xml")
        ]

    @property
    def calibration_hrefs(self) -> List[Tuple[str, str]]:
        return [
            (
                "schema-calibration-{}".format(
                    *extract_properties(x, ["polarisation"])
                ),
                os.path.join(self.granule_href, self.map_filename(x)),
            )
            for x in self.grouped_hrefs["calibration_calibration"]
        ]

    @property
    def noise_hrefs(self) -> List[Tuple[str, str]]:
        return [
            (
                "schema-noise-{}".format(*extract_properties(x, ["polarisation"])),
                os.path.join(self.granule_href, self.map_filename(x)),
            )
            for x in self.grouped_hrefs["calibration_noise"]
        ]

    def create_manifest_asset(self) -> Tuple[str, pystac.asset.Asset]:
        desc = (
            "General product metadata in XML format. Contains a high-level textual "
            "description of the product and references to all of product's components, "
            "the product metadata, including the product identification and the resource "
            "references, and references to the physical location of each component file "
            "contained in the product."
        )

        asset = pystac.Asset(
            href=self.href,
            media_type=pystac.MediaType.XML,
            title="Manifest File",
            roles=["metadata"],
            description=desc,
        )
        return SAFE_MANIFEST_ASSET_KEY, asset

    def create_product_asset(self) -> List[Tuple[str, pystac.asset.Asset]]:
        assets = []
        desc = (
            "Describes the main characteristics corresponding to the band: state of the "
            "platform during acquisition, image properties, Doppler information, geographic "
            "location, etc."
        )
        for key, href in self.annotation_hrefs:
            # Extract polarisation from href
            polarisation = extract_polarisation(href)
            if polarisation:
                # Add polarisation to title
                title = f"{polarisation} Product Schema"
                asset = pystac.Asset(
                    href=href,
                    media_type=pystac.MediaType.XML,
                    title=title,
                    roles=["metadata"],
                    description=desc,
                )
                assets.append((key, asset))
        return assets

    def create_calibration_asset(self) -> List[Tuple[str, pystac.asset.Asset]]:
        assets = []
        desc = (
            "Calibration metadata including calibration information and the beta nought, "
            "sigma nought, gamma and digital number look-up tables that can be used for "
            "absolute product calibration."
        )
        for key, href in self.calibration_hrefs:
            # Extract polarisation from href
            polarisation = extract_polarisation(href)
            if polarisation:
                # Add polarisation to title
                title = f"{polarisation} Calibration Schema"
                asset = pystac.Asset(
                    href=href,
                    media_type=pystac.MediaType.XML,
                    title=title,
                    roles=["metadata"],
                    description=desc,
                )
                assets.append((key, asset))
        return assets

    def create_noise_asset(self) -> List[Tuple[str, pystac.asset.Asset]]:
        assets = []
        for key, href in self.noise_hrefs:
            # Extract polarisation from href
            polarisation = extract_polarisation(href)
            if polarisation:
                # Add polarisation to title
                title = f"{polarisation} Noise Schema"
                asset = pystac.Asset(
                    href=href,
                    media_type=pystac.MediaType.XML,
                    title=title,
                    roles=["metadata"],
                    description="Estimated thermal noise look-up tables",
                )
                assets.append((key, asset))
        return assets



import click

from stactools.sentinel1.grd.commands import grd_cmd
from stactools.sentinel1.rtc.commands import rtc_cmd
from stactools.sentinel1.slc.commands import slc_cmd


def create_sentinel1_command(cli: click.Group) -> click.Command:
    @cli.group("sentinel1")
    def sentinel1_cmd() -> None:
        """Commands for working with sentinel1 data"""
        pass

    sentinel1_cmd.add_command(grd_cmd)
    sentinel1_cmd.add_command(rtc_cmd)
    sentinel1_cmd.add_command(slc_cmd)

    return sentinel1_cmd



from typing import Dict, TypeVar

import pystac
from pystac.extensions.sar import (
    FrequencyBand,
    ObservationDirection,
    Polarization,
    SarExtension,
)
from pystac.extensions.sat import OrbitState, SatExtension
from pystac.utils import str_to_datetime
from stactools.core.io.xml import XmlElement

T = TypeVar("T", pystac.Item, pystac.Asset)


class ProductDataEntry:
    def __init__(
        self,
        resolution_rng: float,
        resolution_azi: float,
        pixel_spacing_rng: float,
        pixel_spacing_azi: float,
        no_looks_rng: int,
        no_looks_azi: int,
        enl: float,
    ):
        self.resolution_rng = resolution_rng
        self.resolution_azi = resolution_azi
        self.pixel_spacing_rng = pixel_spacing_rng
        self.pixel_spacing_azi = pixel_spacing_azi
        self.no_looks_rng = no_looks_rng
        self.no_looks_azi = no_looks_azi
        self.enl = enl


product_data_summary: Dict[str, Dict[str, ProductDataEntry]] = {
    # https://sentinels.copernicus.eu/web/sentinel/technical-guides/sentinel-1-sar/products-algorithms/level-1/single-look-complex/stripmap
    "SM": {
        "S1": ProductDataEntry(1.7, 4.9, 1.5, 3.6, 1, 1, 1),
        "S2": ProductDataEntry(2.0, 4.9, 1.8, 4.2, 1, 1, 1),
        "S3": ProductDataEntry(2.5, 4.9, 2.2, 3.5, 1, 1, 1),
        "S4": ProductDataEntry(3.3, 4.9, 2.6, 4.1, 1, 1, 1),
        "S5": ProductDataEntry(3.3, 3.9, 2.9, 3.6, 1, 1, 1),
        "S6": ProductDataEntry(3.6, 4.9, 3.1, 4.1, 1, 1, 1),
    },
    # https://sentinels.copernicus.eu/web/sentinel/technical-guides/sentinel-1-sar/products-algorithms/level-1/single-look-complex/interferometric-wide-swath
    "IW": {
        "IW1": ProductDataEntry(2.7, 22.5, 2.3, 14.1, 1, 1, 1),
        "IW2": ProductDataEntry(3.1, 22.7, 2.3, 14.1, 1, 1, 1),
        "IW3": ProductDataEntry(3.5, 22.6, 2.3, 14.1, 1, 1, 1),
    },
    # https://sentinels.copernicus.eu/web/sentinel/technical-guides/sentinel-1-sar/products-algorithms/level-1/single-look-complex/extra-wide-swath
    "EW": {
        "EW1": ProductDataEntry(7.9, 43.7, 5.9, 19.9, 1, 1, 1),
        "EW2": ProductDataEntry(9.9, 44.3, 5.9, 19.9, 1, 1, 1),
        "EW3": ProductDataEntry(11.6, 45.2, 5.9, 19.9, 1, 1, 1),
        "EW4": ProductDataEntry(13.3, 45.6, 5.9, 19.9, 1, 1, 1),
        "EW5": ProductDataEntry(14.4, 44.0, 5.9, 19.9, 1, 1, 1),
    },
    # https://sentinels.copernicus.eu/web/sentinel/technical-guides/sentinel-1-sar/products-algorithms/level-1/single-look-complex/wave
    "WV": {
        "WV1": ProductDataEntry(2, 4.8, 1.8, 4.1, 1, 1, 1),
        "WV2": ProductDataEntry(3.1, 4.8, 2.7, 4.1, 1, 1, 1),
    },
}


def fill_common_sar_properties(sar_ext: SarExtension[T], manifest: XmlElement) -> None:
    """Fills the properties for SAR.

    Based on the sar Extension.py

    Args:
        sar_ext (SarExtension): The extension to be populated.
        manifest (XmlElement): manifest.safe file parsed into an XmlElement
    """
    # Fixed properties
    sar_ext.frequency_band = FrequencyBand("C")
    sar_ext.center_frequency = 5.405
    sar_ext.observation_direction = ObservationDirection.RIGHT

    # Read properties
    instrument_mode = manifest.find_text(".//s1sarl1:mode")
    if instrument_mode:
        sar_ext.instrument_mode = instrument_mode
    sar_ext.polarizations = [
        Polarization(x.text)
        for x in manifest.findall(".//s1sarl1:transmitterReceiverPolarisation")
    ]
    product_type = manifest.find_text(".//s1sarl1:productType")
    if product_type:
        sar_ext.product_type = product_type


def fill_swath_sar_properties(
    sar_ext: SarExtension[pystac.Asset], swath: str, polarisation: str
) -> None:
    """Fills the properties for SAR.

    Based on the sar Extension.py

    Args:
        sar_ext (SarExtension): The extension to be populated.
        swath (str): the swath ID for this particular asset
        polarisation (str): the polarisation for this particular asset
    """
    # Properties depending on mode and swath
    product_data = product_data_summary[sar_ext.instrument_mode][swath]

    sar_ext.polarizations = [Polarization(polarisation)]
    sar_ext.resolution_range = product_data.resolution_rng
    sar_ext.resolution_azimuth = product_data.resolution_azi
    sar_ext.pixel_spacing_range = product_data.pixel_spacing_rng
    sar_ext.pixel_spacing_azimuth = product_data.pixel_spacing_azi
    sar_ext.looks_range = product_data.no_looks_rng
    sar_ext.looks_azimuth = product_data.no_looks_azi
    sar_ext.looks_equivalent_number = product_data.enl


def fill_sat_properties(sat_ext: SatExtension[T], manifest: XmlElement) -> None:
    """Fills the properties for SAT.

    Based on the sat Extension.py

    Args:
        sat_ext (SatExtension): The extension to be populated.
        manifest (XmlElement): manifest.safe file parsed into an XmlElement
    """

    sat_ext.platform_international_designator = manifest.find_text(
        ".//safe:nssdcIdentifier"
    )

    orbit_state = manifest.find_text(".//s1:pass")
    if orbit_state:
        sat_ext.orbit_state = OrbitState(orbit_state.lower())

    orbit_number = manifest.find_text(".//safe:orbitNumber")
    if orbit_number:
        sat_ext.absolute_orbit = int(orbit_number)

    relative_orbit = manifest.find_text(".//safe:relativeOrbitNumber")
    if relative_orbit:
        sat_ext.relative_orbit = int(relative_orbit)

    ascending_node_time = manifest.find_text(".//s1:ascendingNodeTime")
    if ascending_node_time:
        sat_ext.anx_datetime = str_to_datetime(ascending_node_time)


def fill_processing_properties(item: pystac.Item, manifest: XmlElement) -> None:
    """Fills the properties for processing.

    Args:
        item (pystac.Item): The extension to be populated.
        manifest (XmlElement): manifest.safe file parsed into an XmlElement
    """
    schema_uri = "https://stac-extensions.github.io/processing/v1.1.0/schema.json"
    if item.stac_extensions is None:
        item.stac_extensions = [schema_uri]
    elif schema_uri not in item.stac_extensions:
        item.stac_extensions.append(schema_uri)

    facility_name = manifest.find_attr("name", ".//safe:facility/")
    if facility_name:
        item.properties["processing:facility"] = facility_name

    # SLCs are Level-1 products
    item.properties["processing:level"] = "L1"

    software = manifest.find(".//safe:software")
    if software is not None:
        name = software.get_attr("name")
        if name:
            item.properties["processing:software"] = {
                name: software.get_attr("version")
            }



from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Tuple

from pystac.utils import str_to_datetime
from shapely.geometry import Polygon, mapping
from stactools.core.io import ReadHrefModifier
from stactools.core.io.xml import XmlElement

from .metadata_links import MetadataLinks


class ProductMetadataError(Exception):
    pass


def get_shape(
    meta_links: MetadataLinks,
    read_href_modifier: Optional[ReadHrefModifier],
    **kwargs: Any,
) -> List[int]:
    links = meta_links.create_product_asset()
    root = XmlElement.from_file(links[0][1].href, read_href_modifier, **kwargs)

    num_samples = root.find_text(".//numberOfSamples")
    num_lines = root.find_text(".//numberOfLines")

    if num_samples and num_lines:
        x_size = int(num_samples)
        y_size = int(num_lines)
        return [x_size, y_size]

    raise ValueError(
        "Cannot determine shape, samples and lines, using product metadata "
        f"in {links[0][1].href}"
    )


class ProductMetadata:
    def __init__(
        self,
        href: str,
        file_hrefs: Dict[str, List[str]],
        file_mapper: Callable[[str], str],
        manifest: XmlElement,
    ) -> None:
        self.href = href
        self._root = manifest
        self.file_hrefs = file_hrefs
        self.file_mapper = file_mapper

        def _get_geometries() -> Tuple[List[float], Dict[str, Any]]:
            # Find the footprint descriptor
            footprint_text = self._root.find_text(".//gml:coordinates")
            if footprint_text is None:
                raise ProductMetadataError(
                    f"Cannot parse footprint from product metadata at {self.href}"
                )

            # Convert to values
            footprint_value = [
                float(x) for x in footprint_text.replace(" ", ",").split(",")
            ]

            footprint_points = [
                p[::-1] for p in list(zip(*[iter(footprint_value)] * 2))
            ]

            footprint_polygon = Polygon(footprint_points)
            geometry = mapping(footprint_polygon)
            bbox = list(footprint_polygon.bounds)

            return (bbox, geometry)

        self.bbox, self.geometry = _get_geometries()

    @property
    def scene_id(self) -> str:
        """Returns the string to be used for a STAC Item id.

        Removes the processing number and .SAFE extension
        from the product_id defined below.

        Parsed based on the naming convention found here:
        https://sentinel.esa.int/web/sentinel/user-guides/sentinel-2-msi/naming-convention
        """
        scene_id = self.product_id.split(".")[0]

        return scene_id

    @property
    def product_id(self) -> str:
        # Parse the name from href as it doesn't exist in xml files
        href = self.href
        result = href.split("/")[-2]
        if result is None:
            raise ValueError(
                "Cannot determine product ID using product metadata " f"at {self.href}"
            )
        else:
            return result

    @property
    def get_datetime(self) -> datetime:
        start_time = self._root.find_text(".//safe:startTime")
        end_time = self._root.find_text(".//safe:stopTime")

        if start_time is not None:
            central_time = (
                datetime.strptime(str(start_time), "%Y-%m-%dT%H:%M:%S.%f")
                + (
                    datetime.strptime(str(end_time), "%Y-%m-%dT%H:%M:%S.%f")
                    - datetime.strptime(str(start_time), "%Y-%m-%dT%H:%M:%S.%f")
                )
                / 2
            )

        if central_time is None:
            raise ValueError(
                "Cannot determine product start time using product metadata "
                f"at {self.href}"
            )
        else:
            return str_to_datetime(str(central_time))

    @property
    def start_datetime(self) -> datetime:
        time = self._root.find_text(".//safe:startTime")

        if time is None:
            raise ValueError(
                "Cannot determine product start time using product metadata "
                f"at {self.href}"
            )
        else:
            return str_to_datetime(f"{time}Z")

    @property
    def end_datetime(self) -> datetime:
        time = self._root.findall(".//safe:stopTime")

        if time is None:
            raise ValueError(
                "Cannot determine product start time using product metadata "
                f"at {self.href}"
            )
        else:
            return str_to_datetime(f"{time[0].text}Z")

    @property
    def platform(self) -> Optional[str]:

        family_name = self._root.findall(".//safe:familyName")[0].text
        assert family_name is not None
        platform_name = self._root.findall(".//safe:number")[0].text
        assert platform_name is not None

        return f"{family_name}{platform_name}"

    @property
    def cycle_number(self) -> Optional[str]:

        return self._root.findall(".//safe:cycleNumber")[0].text

    @property
    def image_paths(self) -> List[str]:
        return [self.file_mapper(x) for x in self.file_hrefs["measurement"]]

    def orbit_source(self) -> str:
        for resource in self._root.findall(
            ".//{http://www.esa.int/safe/sentinel-1.0}resource[@role]"
        ):

            name = resource.find_attr("name", ".")
            if name is None or not name.endswith(".EOF"):
                continue

            role = resource.find_attr("role", ".")
            if role is None or not role.startswith("AUX_"):
                continue

            if role == "AUX_POE":
                return "POEORB"
            elif role == "AUX_RES":
                return "RESORB"
            elif role == "AUX_PRE":
                return "PREORB"
            else:
                raise RuntimeError(f"Invalid orbit file role found: {role}")

        return "DOWNLINK"



from typing import Any, Callable, Dict, List

from pystac.utils import datetime_to_str
from stactools.core.io.xml import XmlElement

from ..product_metadata import ProductMetadata


class GRDProductMetadata(ProductMetadata):
    def __init__(
        self,
        href: str,
        file_hrefs: Dict[str, List[str]],
        file_mapper: Callable[[str], str],
        manifest: XmlElement,
    ) -> None:
        super().__init__(href, file_hrefs, file_mapper, manifest)
        self.resolution = self.product_id.split("_")[2][-1]

    @property
    def metadata_dict(self) -> Dict[str, Any]:

        resolutions = {"F": "full", "H": "high", "M": "medium"}

        tmp = self._root.find(".//s1sarl1:sliceNumber")
        slice_number = tmp.text if tmp is not None else None
        tmp = self._root.find(".//s1sarl1:totalSlices")
        total_slices = tmp.text if tmp is not None else None

        result = {
            "start_datetime": datetime_to_str(self.start_datetime),
            "end_datetime": datetime_to_str(self.end_datetime),
            "s1:instrument_configuration_ID": self._root.find_text(
                ".//s1sarl1:instrumentConfigurationID"
            ),
            "s1:datatake_id": self._root.find_text(".//s1sarl1:missionDataTakeID"),
            "s1:product_timeliness": self._root.find_text(
                ".//s1sarl1:productTimelinessCategory"
            ),
            "s1:processing_level": self.product_id.split("_")[3][0],
            "s1:resolution": resolutions[self.resolution],
            "s1:orbit_source": self.orbit_source(),
            "s1:slice_number": slice_number,
            "s1:total_slices": total_slices,
        }

        return {k: v for k, v in result.items() if v is not None}



from typing import Any, Dict

from pystac.utils import datetime_to_str

from ..product_metadata import ProductMetadata


class SLCProductMetadata(ProductMetadata):
    @property
    def metadata_dict(self) -> Dict[str, Any]:
        result = {
            "start_datetime": datetime_to_str(self.start_datetime),
            "end_datetime": datetime_to_str(self.end_datetime),
            "s1:instrument_configuration_ID": self._root.find_text(
                ".//s1sarl1:instrumentConfigurationID"
            ),
            "s1:datatake_id": self._root.find_text(".//s1sarl1:missionDataTakeID"),
            "s1:product_timeliness": self._root.find_text(
                ".//s1sarl1:productTimelinessCategory"
            ),
            "s1:orbit_source": self.orbit_source(),
            "s1:slice_number": self._root.find_text(".//s1sarl1:sliceNumber"),
            "s1:total_slices": self._root.find_text(".//s1sarl1:totalSlices"),
            "s1:swaths": [elem.text for elem in self._root.findall(".//s1sarl1:swath")],
        }

        return {k: v for k, v in result.items() if v is not None}



import pystac
from pystac import ProviderRole
from pystac.extensions.eo import Band
from pystac.link import Link

INSPIRE_METADATA_ASSET_KEY = "inspire-metadata"
SAFE_MANIFEST_ASSET_KEY = "safe-manifest"
PRODUCT_METADATA_ASSET_KEY = "product-metadata"

SENTINEL_LICENSE = Link(
    rel="license",
    target="https://sentinel.esa.int/documents/"
    + "247904/690755/Sentinel_Data_Legal_Notice",
)

SENTINEL_PLATFORMS = ["sentinel-1a", "sentinel-1b"]

ACQUISITION_MODES = [
    "Stripmap (SM)",
    "Interferometric Wide Swath (IW)",
    "Extra Wide Swath (EW)",
    "Wave (WV)",
]
SENTINEL_CONSTELLATION = "sentinel-1"


SENTINEL_PROVIDER = pystac.Provider(
    name="ESA",
    roles=[
        ProviderRole.PRODUCER,
        ProviderRole.PROCESSOR,
        ProviderRole.LICENSOR,
    ],
    url="https://earth.esa.int/eogateway",
)


SENTINEL_POLARIZATIONS = {
    "vh": Band.create(
        name="VH",
        description="VH band: vertical transmit and horizontal receive",
    ),
    "hh": Band.create(
        name="HH",
        description="HH band: horizontal transmit and horizontal receive",
    ),
    "hv": Band.create(
        name="HV",
        description="HV band: horizontal transmit and vertical receive",
    ),
    "vv": Band.create(
        name="VV",
        description="VV band: vertical transmit and vertical receive",
    ),
}



from datetime import datetime
from typing import Any, Dict

import pystac
from pystac import Extent, ProviderRole, SpatialExtent, TemporalExtent
from pystac.extensions import sar, sat
from pystac.extensions.item_assets import AssetDefinition
from pystac.link import Link
from pystac.utils import str_to_datetime

from ..constants import (
    ACQUISITION_MODES,
    INSPIRE_METADATA_ASSET_KEY,
    PRODUCT_METADATA_ASSET_KEY,
    SAFE_MANIFEST_ASSET_KEY,
    SENTINEL_CONSTELLATION,
    SENTINEL_LICENSE,
    SENTINEL_PLATFORMS,
    SENTINEL_POLARIZATIONS,
    SENTINEL_PROVIDER,
)

__all__ = [
    "INSPIRE_METADATA_ASSET_KEY",
    "SAFE_MANIFEST_ASSET_KEY",
    "PRODUCT_METADATA_ASSET_KEY",
    "SENTINEL_LICENSE",
    "SENTINEL_PLATFORMS",
    "ACQUISITION_MODES",
    "SENTINEL_CONSTELLATION",
    "SENTINEL_PROVIDER",
    "SENTINEL_POLARIZATIONS",
    "SENTINEL_GRD_DESCRIPTION",
    "SENTINEL_GRD_START",
    "SENTINEL_GRD_EXTENT",
    "SENTINEL_GRD_PROVIDER",
    "SENTINEL_GRD_TECHNICAL_GUIDE",
    "SENTINEL_GRD_LICENSE",
    "SENTINEL_GRD_KEYWORDS",
    "SENTINEL_GRD_SAT",
    "SENTINEL_GRD_SAR",
    "SENTINEL_GRD_ASSETS",
]


SENTINEL_GRD_DESCRIPTION = (
    "Level-1 Ground Range Detected (GRD) products consist of focused SAR data that has been detected, multi-looked and projected to ground range using an Earth ellipsoid model. The ellipsoid projection of the GRD products is corrected using the terrain height specified in the product general annotation. "  # noqa: E501
    "The terrain height used varies in azimuth but is constant in range. Ground range coordinates are the slant range coordinates projected onto the ellipsoid of the Earth. Pixel values represent detected magnitude. Phase information is lost. The resulting product has approximately square spatial resolution "  # noqa: E501
    "and square pixel spacing with reduced speckle due to the multi-look processing. The noise vector annotation data set, within the product annotations, contains thermal noise vectors so that users can apply a thermal noise correction by subtracting the noise from the power detected image. "  # noqa: E501
    "The thermal noise correction is, for example, supported by the Sentinel-1 Toolbox (S1TBX). For the IW and EW GRD products, multi-looking is performed on each burst individually. All bursts in all sub-swaths are then seamlessly merged to form a single, contiguous, ground range detected image per polarization channel."  # noqa: E501
    "GRD products are available in three resolutions, characterised by the acquisition mode and the level of multi-looking applied: Full Resolution (FR), High Resolution (HR), Medium Resolution (MR)."  # noqa: E501
)

SENTINEL_GRD_START: datetime = str_to_datetime("2014-10-10T00:00:00Z")
SENTINEL_GRD_EXTENT = Extent(
    SpatialExtent([-180.0, -90.0, 180.0, 90.0]),
    TemporalExtent([[SENTINEL_GRD_START, None]]),
)

SENTINEL_GRD_PROVIDER = pystac.Provider(
    name="Sinergise",
    roles=[
        ProviderRole.HOST,
        ProviderRole.PROCESSOR,
        ProviderRole.LICENSOR,
    ],
    url="https://registry.opendata.aws/sentinel-1/",
)

SENTINEL_GRD_TECHNICAL_GUIDE = Link(
    title="Sentinel-1 Ground Range Detected (GRD) Technical Guide",
    rel="about",
    target="https://sentinels.copernicus.eu/web/sentinel/technical-guides/sentinel-1-sar/products-algorithms/level-1-algorithms/ground-range-detected",  # noqa: E501
)

SENTINEL_GRD_LICENSE = Link(
    title="Sentinel License",
    rel="license",
    target="https://scihub.copernicus.eu/twiki/do/view/SciHubWebPortal/TermsConditions",
)

SENTINEL_GRD_KEYWORDS = ["ground", "sentinel", "copernicus", "esa", "sar"]

SENTINEL_GRD_SAT = {
    "orbit_state": [sat.OrbitState.ASCENDING, sat.OrbitState.DESCENDING]
}

SENTINEL_GRD_SAR: Dict[str, Any] = {
    "looks_range": [2, 3, 5, 6],
    "product_type": ["GRD"],
    "looks_azimuth": [1, 2, 6],
    "polarizations": [
        sar.Polarization.HH,
        sar.Polarization.VV,
        sar.Polarization.HV,
        sar.Polarization.VH,
        [
            sar.Polarization.HH,
            sar.Polarization.HV,
        ],
        [
            sar.Polarization.VV,
            sar.Polarization.VH,
        ],
    ],
    "frequency_band": [sar.FrequencyBand.C],
    "instrument_mode": ["IW", "EW", "SM"],
    "center_frequency": [5.405],
    "resolution_range": [9, 20, 23, 50, 93],
    "resolution_azimuth": [9, 22, 23, 50, 87],
    "pixel_spacing_range": [3.5, 10, 25, 40],
    "observation_direction": [sar.ObservationDirection.RIGHT],
    "pixel_spacing_azimuth": [3.5, 10, 25, 40],
    "looks_equivalent_number": [3.7, 29.7, 398.4, 4.4, 81.8, 2.8, 10.7, 123.7],
}

SENTINEL_GRD_ASSETS = {
    "vh": AssetDefinition(
        {
            "title": "VH Data",
            "type": pystac.MediaType.COG,
            "description": "VH polarization backscattering coefficient, 16-bit DN.",
            "roles": ["data"],
        }
    ),
    "hh": AssetDefinition(
        {
            "title": "HH Data",
            "type": pystac.MediaType.COG,
            "description": "HH polarization backscattering coefficient, 16-bit DN.",
            "roles": ["data"],
        }
    ),
    "hv": AssetDefinition(
        {
            "title": "HV Data",
            "type": pystac.MediaType.COG,
            "description": "HV polarization backscattering coefficient, 16-bit DN.",
            "roles": ["data"],
        }
    ),
    "vv": AssetDefinition(
        {
            "title": "VV Data",
            "type": pystac.MediaType.COG,
            "description": "VV polarization backscattering coefficient, 16-bit DN.",
            "roles": ["data"],
        }
    ),
    "schema-calibration-hh": AssetDefinition(
        {
            "title": "HH Calibration Schema",
            "type": pystac.MediaType.XML,
            "description": (
                "Calibration metadata including calibration information and the beta nought, "
                "sigma nought, gamma and digital number look-up tables that can be used for "
                "absolute product calibration."
            ),
            "roles": ["metadata"],
        }
    ),
    "schema-calibration-hv": AssetDefinition(
        {
            "title": "HV Calibration Schema",
            "type": pystac.MediaType.XML,
            "description": (
                "Calibration metadata including calibration information and the beta nought, "
                "sigma nought, gamma and digital number look-up tables that can be used for "
                "absolute product calibration."
            ),
            "roles": ["metadata"],
        }
    ),
    "schema-calibration-vh": AssetDefinition(
        {
            "title": "VH Calibration Schema",
            "type": pystac.MediaType.XML,
            "description": (
                "Calibration metadata including calibration information and the beta nought, "
                "sigma nought, gamma and digital number look-up tables that can be used for "
                "absolute product calibration."
            ),
            "roles": ["metadata"],
        }
    ),
    "schema-calibration-vv": AssetDefinition(
        {
            "title": "VV Calibration Schema",
            "type": pystac.MediaType.XML,
            "description": (
                "Calibration metadata including calibration information and the beta nought, "
                "sigma nought, gamma and digital number look-up tables that can be used for "
                "absolute product calibration."
            ),
            "roles": ["metadata"],
        }
    ),
    "schema-noise-hh": AssetDefinition(
        {
            "title": "HH Noise Schema",
            "type": pystac.MediaType.XML,
            "description": "Estimated thermal noise look-up tables",
            "roles": ["metadata"],
        }
    ),
    "schema-noise-hv": AssetDefinition(
        {
            "title": "HV Noise Schema",
            "type": pystac.MediaType.XML,
            "description": "Estimated thermal noise look-up tables",
            "roles": ["metadata"],
        }
    ),
    "schema-noise-vh": AssetDefinition(
        {
            "title": "VH Noise Schema",
            "type": pystac.MediaType.XML,
            "description": "Estimated thermal noise look-up tables",
            "roles": ["metadata"],
        }
    ),
    "schema-noise-vv": AssetDefinition(
        {
            "title": "VV Noise Schema",
            "type": pystac.MediaType.XML,
            "description": "Estimated thermal noise look-up tables",
            "roles": ["metadata"],
        }
    ),
    "schema-product-hh": AssetDefinition(
        {
            "title": "HH Product Schema",
            "type": pystac.MediaType.XML,
            "description": (
                "Describes the main characteristics corresponding to the band: state of the "
                "platform during acquisition, image properties, Doppler information, geographic "
                "location, etc."
            ),
            "roles": ["metadata"],
        }
    ),
    "schema-product-hv": AssetDefinition(
        {
            "title": "HV Product Schema",
            "type": pystac.MediaType.XML,
            "description": (
                "Describes the main characteristics corresponding to the band: state of the "
                "platform during acquisition, image properties, Doppler information, geographic "
                "location, etc."
            ),
            "roles": ["metadata"],
        }
    ),
    "schema-product-vh": AssetDefinition(
        {
            "title": "VH Product Schema",
            "type": pystac.MediaType.XML,
            "description": (
                "Describes the main characteristics corresponding to the band: state of the "
                "platform during acquisition, image properties, Doppler information, geographic "
                "location, etc."
            ),
            "roles": ["metadata"],
        }
    ),
    "schema-product-vv": AssetDefinition(
        {
            "title": "VV Product Schema",
            "type": pystac.MediaType.XML,
            "description": (
                "Describes the main characteristics corresponding to the band: state of the "
                "platform during acquisition, image properties, Doppler information, geographic "
                "location, etc."
            ),
            "roles": ["metadata"],
        }
    ),
    "safe-manifest": AssetDefinition(
        {
            "title": "Manifest File",
            "type": pystac.MediaType.XML,
            "description": (
                "General product metadata in XML format. Contains a high-level textual "
                "description of the product and references to all of product's components, "
                "the product metadata, including the product identification and the resource "
                "references, and references to the physical location of each component file "
                "contained in the product."
            ),
            "roles": ["metadata"],
        }
    ),
    "thumbnail": AssetDefinition(
        {
            "title": "Preview Image",
            "type": pystac.MediaType.PNG,
            "description": (
                "An averaged, decimated preview image in PNG format. Single polarization "
                "products are represented with a grey scale image. Dual polarization products "
                "are represented by a single composite colour image in RGB with the red channel "
                "(R) representing the  co-polarization VV or HH), the green channel (G) "
                "represents the cross-polarization (VH or HV) and the blue channel (B) "
                "represents the ratio of the cross an co-polarizations."
            ),
            "roles": ["thumbnail"],
        }
    ),
}



from datetime import datetime
from typing import Any, Dict

import pystac
from pystac import Extent, ProviderRole, SpatialExtent, TemporalExtent
from pystac.extensions import sar
from pystac.link import Link
from pystac.utils import str_to_datetime

# General Sentinel-1 Constants
# -
SENTINEL_LICENSE = Link(
    rel="license",
    target="https://sentinel.esa.int/documents/"
    + "247904/690755/Sentinel_Data_Legal_Notice",
)

SENTINEL_INSTRUMENTS = ["c-sar"]
SENTINEL_CONSTELLATION = "sentinel-1"
SENTINEL_PLATFORMS = ["sentinel-1a", "sentinel-1b"]
SENTINEL_FREQUENCY_BAND = sar.FrequencyBand.C
SENTINEL_CENTER_FREQUENCY = 5.405
SENTINEL_OBSERVATION_DIRECTION = sar.ObservationDirection.RIGHT

SENTINEL_PROVIDER = pystac.Provider(
    name="ESA",
    roles=[ProviderRole.LICENSOR, ProviderRole.PRODUCER],
    url="https://sentinel.esa.int/web/sentinel/missions/sentinel-1",
)

SENTINEL_LICENSE = Link(
    rel="license", target="https://spacedata.copernicus.eu/data-offer/legal-documents"
)

# RTC-specific constants
# -
SENTINEL_RTC_PROVIDER = pystac.Provider(
    name="Indigo Ag Inc.",
    roles=[ProviderRole.LICENSOR, ProviderRole.PROCESSOR, ProviderRole.HOST],
    url="https://registry.opendata.aws/sentinel-1-rtc-indigo",
    extra_fields={
        "processing:level": "L3",
        "processing:lineage": "https://sentinel-s1-rtc-indigo-docs.s3-us-west-2.amazonaws.com/methodology.html",  # noqa: E501
        "processing:software": {"S1TBX": "7.0.2"},
    },
)

SENTINEL_RTC_LICENSE = Link(
    rel="license", target="https://www.indigoag.com/forms/atlas-sentinel-license"
)

SENTINEL_RTC_DESCRIPTION = "Sentinel1 radiometric terrain corrected backscatter (RTC) over CONUS. The Sentinel-1 mission is a constellation of C-band Synthetic Aperture Radar (SAR) satellites from the European Space Agency launched since 2014. These satellites collect observations of radar backscatter intensity day or night, regardless of the weather conditions, making them enormously valuable for environmental monitoring. These radar data have been processed from original Ground Range Detected (GRD) scenes into a Radiometrically Terrain Corrected, tiled product suitable for analysis. This product is available over the Contiguous United States (CONUS) since 2017 when Sentinel-1 data became globally available."  # noqa: E501

SENTINEL_RTC_START: datetime = str_to_datetime("2016-07-29T00:00:00Z")
SENTINEL_RTC_EXTENT = Extent(
    SpatialExtent([-124.73460, 24.54254, -66.89191, 49.36949]),
    TemporalExtent([[SENTINEL_RTC_START, None]]),
)

utm_zones = ["10", "11", "12", "13", "14", "15", "16", "17", "18", "19"]
SENTINEL_RTC_EPSGS = [int(f"326{x}") for x in utm_zones]

# RTC is derived from S1 GRD, so include input GRD properties for IW2 (center swath)
# https://sentinel.esa.int/web/sentinel/technical-guides/sentinel-1-sar/products-algorithms/level-1-algorithms/ground-range-detected/iw
SENTINEL_RTC_SAR: Dict[str, Any] = {
    "instrument_mode": "IW",
    "product_type": "RTC",
    "polarizations": [sar.Polarization.VV, sar.Polarization.VH],
    "resolution_range": 20.3,
    "resolution_azimuth": 22.6,
    "pixel_spacing_range": 10,
    "pixel_spacing_azimuth": 10,
    "looks_equivalent_number": 4.3,
    "looks_range": 5,
    "looks_azimuth": 1,
    "gsd": 20,  # final MGRS pixel posting
}



from typing import Dict, TypeVar

import pystac
from pystac.extensions.sar import (
    FrequencyBand,
    ObservationDirection,
    Polarization,
    SarExtension,
)
from pystac.extensions.sat import OrbitState, SatExtension
from stactools.core.io.xml import XmlElement

T = TypeVar("T", pystac.Item, pystac.Asset)


class ProductDataEntry:
    def __init__(
        self,
        resolution_rng: float,
        resolution_azi: float,
        pixel_spacing_rng: float,
        pixel_spacing_azi: float,
        no_looks_rng: int,
        no_looks_azi: int,
        enl: float,
    ):
        self.resolution_rng = resolution_rng
        self.resolution_azi = resolution_azi
        self.pixel_spacing_rng = pixel_spacing_rng
        self.pixel_spacing_azi = pixel_spacing_azi
        self.no_looks_rng = no_looks_rng
        self.no_looks_azi = no_looks_azi
        self.enl = enl


# Sourced from Sentinel-1 Product Definition: Table 5-1
#   https://sentinel.esa.int/web/sentinel/user-guides/sentinel-1-sar/document-library/-/asset_publisher/1dO7RF5fJMbd/content/sentinel-1-product-definition
product_data_summary: Dict[str, Dict[str, ProductDataEntry]] = {
    "SM": {
        "F": ProductDataEntry(9, 9, 3.5, 3.5, 2, 2, 3.7),
        "H": ProductDataEntry(23, 23, 10, 10, 6, 6, 29.7),
        "M": ProductDataEntry(84, 84, 40, 40, 22, 22, 398.4),
    },
    "IW": {
        "H": ProductDataEntry(20, 22, 10, 10, 5, 1, 4.4),
        "M": ProductDataEntry(88, 87, 40, 40, 22, 5, 81.8),
    },
    "EW": {
        "H": ProductDataEntry(50, 50, 25, 25, 3, 1, 2.7),
        "M": ProductDataEntry(93, 87, 40, 40, 6, 2, 10.7),
    },
    "WV": {"M": ProductDataEntry(52, 51, 25, 25, 13, 13, 123.7)},
}


def fill_sar_properties(
    sar_ext: SarExtension[T], manifest: XmlElement, resolution: str
) -> None:
    """Fills the properties for SAR.

    Based on the sar Extension.py

    Args:
        sar_ext (SarExtension): The extension to be populated.
        resolution (str): product resolution, needed to select metadata from
            static values in product_data_summary
        manifest (XmlElement): manifest.safe file parsed into an XmlElement
    """
    # Fixed properties
    sar_ext.frequency_band = FrequencyBand("C")
    sar_ext.center_frequency = 5.405
    sar_ext.observation_direction = ObservationDirection.RIGHT

    # Read properties
    instrument_mode = manifest.find_text(".//s1sarl1:mode")
    if instrument_mode:
        sar_ext.instrument_mode = instrument_mode
    sar_ext.polarizations = [
        Polarization(x.text)
        for x in manifest.findall(".//s1sarl1:transmitterReceiverPolarisation")
    ]
    product_type = manifest.find_text(".//s1sarl1:productType")
    if product_type:
        sar_ext.product_type = product_type

    # Properties depending on mode and resolution
    product_data = product_data_summary[sar_ext.instrument_mode][resolution]

    sar_ext.resolution_range = product_data.resolution_rng
    sar_ext.resolution_azimuth = product_data.resolution_azi
    sar_ext.pixel_spacing_range = product_data.pixel_spacing_rng
    sar_ext.pixel_spacing_azimuth = product_data.pixel_spacing_azi
    sar_ext.looks_range = product_data.no_looks_rng
    sar_ext.looks_azimuth = product_data.no_looks_azi
    sar_ext.looks_equivalent_number = product_data.enl


def fill_sat_properties(sat_ext: SatExtension[T], manifest: XmlElement) -> None:
    """Fills the properties for SAT.

    Based on the sat Extension.py

    Args:
        sat_ext (SatExtension): The extension to be populated.
        manifest (XmlElement): manifest.safe file parsed into an XmlElement
    """

    sat_ext.platform_international_designator = manifest.find_text(
        ".//safe:nssdcIdentifier"
    )

    orbit_state = manifest.find_text(".//s1:pass")
    if orbit_state:
        sat_ext.orbit_state = OrbitState(orbit_state.lower())

    orbit_number = manifest.find_text(".//safe:orbitNumber")
    if orbit_number:
        sat_ext.absolute_orbit = int(orbit_number)

    relative_orbit = manifest.find_text(".//safe:relativeOrbitNumber")
    if relative_orbit:
        sat_ext.relative_orbit = int(relative_orbit)
